package com.persagy.common;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.util.StrUtil;
import com.alibaba.fastjson.JSON;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonInclude;
import io.swagger.annotations.ApiModel;
import io.swagger.annotations.ApiModelProperty;
import lombok.AllArgsConstructor;
import lombok.Data;
import lombok.NoArgsConstructor;
import org.springframework.util.Assert;

import java.io.Serializable;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Objects;

/**
 * 通用返回
 *
 * @param <T> 数据泛型
 */
@Data
@NoArgsConstructor
@AllArgsConstructor
@ApiModel(value = "接口返回对象", description = "接口返回对象")
public class R<T> implements Serializable {
    private static final long serialVersionUID = 1L;
    /**
     * 成功
     */
    private static final String SUCCESS = "success";
    /**
     * 失败
     */
    private static final String FAIL = "fail";

    /**
     * 错误码
     *
     * @see ErrorCode#getCode()
     */
    @ApiModelProperty(value = "结果码")
    @JsonInclude(JsonInclude.Include.NON_NULL)
    private String code;
    /**
     * success、fail
     */
    @ApiModelProperty(value = "结果标识:成功标识：success，失败标识：fail")
    @JsonInclude(JsonInclude.Include.NON_NULL)
    private String result;
    /**
     * 返回数据
     */
    @ApiModelProperty(value = "响应数据")
    @JsonInclude(JsonInclude.Include.NON_NULL)
    private T data;
    /**
     * 错误提示，用户可阅读
     *
     * @see ErrorCode#getMessage() ()
     */
    @ApiModelProperty(value = "错误提示信息")
    @JsonInclude(JsonInclude.Include.NON_NULL)
    private String message;
    /**
     * 返回记录条数或者操作记录条数
     */
    @ApiModelProperty(value = "返回记录条数或者操作记录条数")
    @JsonInclude(JsonInclude.Include.NON_NULL)
    private Integer count;
    /**
     * 记录总条数
     */
    @ApiModelProperty(value = "记录总条数")
    @JsonInclude(JsonInclude.Include.NON_NULL)
    private Long total;
    /**
     * 时间戳
     */
    @ApiModelProperty(value = "响应时间戳")
    private long timestamp = System.currentTimeMillis();

    /**
     * 附加数据
     */
    @JsonInclude(JsonInclude.Include.NON_NULL)
    @ApiModelProperty(value = "附加数据")
    private Map<Object, Object> extra;

    /**
     * 链路traceId
     */
    @JsonInclude(JsonInclude.Include.NON_NULL)
    @ApiModelProperty(value = "链路traceId")
    private String traceId;

    /**
     * 通过 withColumns 可以控制返回的数据内容,可以一定程度上提供服务性能.
     * withColumns 粒度比较粗,一般由服务提供者决定哪些字段可以通过 withColumns 来控制.
     */
    @JsonInclude(JsonInclude.Include.NON_NULL)
    @ApiModelProperty(value = "扩展字段数据")
    private List<String> withColumns;

    /**
     * 通过 includeColumns指定返回字段,可以一定程度上提供服务性能.
     * 由服务提供者过滤字段，提高接口效率
     */
    @JsonInclude(JsonInclude.Include.NON_NULL)
    @ApiModelProperty(value = "指定返回字段")
    private List<String> includeColumns;

    /**
     * 系统报错时，抛出的原生信息
     */
    @ApiModelProperty(value = "异常消息")
    @JsonInclude(JsonInclude.Include.NON_NULL)
    private String cause;
    /**
     * 是否执行默认操作
     */
    @JsonIgnore
    private Boolean defExec = true;

    /**
     * 将传入的 result 对象，转换成另外一个泛型结果的对象
     * <p>
     * 因为 A 方法返回的 R 对象，不满足调用其的 B 方法的返回，所以需要进行转换。
     *
     * @param result 传入的 result 对象
     * @param <T>    返回的泛型
     * @return 新的 R 对象
     */
    public static <T> R<T> fail(R<?> result) {
        return fail(result.getCode(), result.getMessage());
    }


    public static <T> R<T> fail(String code, String message, String cause) {
        Assert.isTrue(!ErrorCodeConstants.SUCCESS.getCode().equals(code), "code 必须是错误的！");
        R<T> result = new R<>();
        result.code = code;
        result.message = message;
        result.cause = cause;
        result.result = FAIL;
        return result;
    }

    public static <T> R<T> fail(String code, String message) {
        return fail(code, message, null);
    }

    public static <T> R<T> fail(ErrorCode errorCode) {
        return fail(errorCode.getCode(), errorCode.getMessage());
    }

    public static <T> R<T> success(T data) {
        R<T> result = new R<>();
        result.code = ErrorCodeConstants.SUCCESS.getCode();
        result.data = data;
        result.result = SUCCESS;
        return result;
    }

    public static boolean isSuccess(String code) {
        return Objects.equals(code, ErrorCodeConstants.SUCCESS.getCode());
    }

    public static <T> R<T> fail(BusinessException businessException) {
        return fail(businessException.getCode(), businessException.getMessage());
    }

    /**
     * 多个数据
     *
     * @param records
     * @param total
     * @param <T>
     * @return
     */
    public static <T> R<Collection<T>> successMulti(Collection<T> records, Long total) {
        R<Collection<T>> result = success(records);
        result.setCount(CollUtil.size(records));
        result.setTotal(total);
        return result;
    }

    /**
     * 多个数据
     *
     * @param records
     * @param <T>
     * @return
     */
    public static <T> R<Collection<T>> successMulti(Collection<T> records) {
        return successMulti(records, null);
    }


    public static <T> R<T> validFail(String msg) {
        return R.fail(ErrorCodeConstants.USER_ERROR_A0421.getCode(), (msg == null || msg.isEmpty()) ? ErrorCodeConstants.USER_ERROR_A0421.getMessage() : msg);
    }

    public static <T> R<T> validFail(String msg, Object... args) {
        String message = (msg == null || msg.isEmpty()) ? ErrorCodeConstants.USER_ERROR_A0421.getMessage() : msg;
        return R.fail(ErrorCodeConstants.USER_ERROR_A0421.getCode(), StrUtil.format(message, args));
    }

    @JsonIgnore // 避免 jackson 序列化
    public boolean getIsFail() {
        return !isSuccess(code);
    }

    @JsonIgnore // 避免 jackson 序列化
    public boolean getIsSuccess() {
        return isSuccess(code);
    }

    /**
     * 判断是否有异常。如果有，则抛出 {@link BusinessException} 异常
     */
    public void checkError() throws BusinessException {
        if (isSuccess(code)) {
            return;
        }
        // 业务异常
        throw new BusinessException(code, message);
    }

    public R<T> setCode(String code) {
        this.code = code;
        return this;
    }

    public R<T> setResult(String result) {
        this.result = result;
        return this;
    }

    public R<T> setData(T data) {
        this.data = data;
        return this;
    }

    public R<T> setMessage(String message) {
        this.message = message;
        return this;
    }

    public R<T> setCount(Integer count) {
        this.count = count;
        return this;
    }

    public R<T> setTotal(Long total) {
        this.total = total;
        return this;
    }

    public R<T> setExtra(Map<Object, Object> extra) {
        this.extra = extra;
        return this;
    }

    public R<T> setCause(String cause) {
        this.cause = cause;
        return this;
    }

    public R<T> setTraceId(String traceId) {
        this.traceId = traceId;
        return this;
    }

    @Override
    public String toString() {
        return JSON.toJSONString(this);
    }
}
