package com.persagy.excel;

import cn.hutool.core.thread.ThreadUtil;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import com.alibaba.excel.context.AnalysisContext;
import com.alibaba.excel.event.AnalysisEventListener;
import com.alibaba.excel.exception.ExcelDataConvertException;
import com.alibaba.excel.metadata.CellExtra;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.TypeReference;
import com.alibaba.fastjson.serializer.SerializerFeature;
import com.persagy.util.DateUtil;
import lombok.*;
import lombok.extern.slf4j.Slf4j;

import java.time.LocalDateTime;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;
import java.util.function.Consumer;

/***
 *
 * 单元格合并参考连接:
 * https://developer.aliyun.com/article/852802
 * https://blog.csdn.net/qq_40823831/article/details/120553658
 * @author lgy
 * @date 2022/4/27 10:15
 */
@Getter
@Setter
@Slf4j
@NoArgsConstructor
@Builder
@AllArgsConstructor
public class ExcelListener<T> extends AnalysisEventListener<T> {

    /**
     * 数据列
     */
    private List<List<String>> dataList = new ArrayList<>();
    /**
     * 表号列
     */
    private List<String> meterList = new ArrayList<>();
    /**
     * 功能号列表
     */
    private List<String> funcidList = new ArrayList<>();
    /**
     * 描述列表
     */
    private List<String> descriptionList = new ArrayList<>();
    /**
     * 根据sheetName读取项目ID
     */
    private String buildding;
    /**
     * 数据处理逻辑
     * 发数入库
     */
    private Consumer<ExcelData> consumer;
    /**
     * 每次批量处理的条数
     */
    private Integer batchSize;
    /**
     * 控制是否继续读取
     */
    private Boolean hasNext = true;
    /**
     * 是否只读取前3行的表头
     */
    private Boolean onlyReadHeader = false;

    /**
     * 是否动态导入
     * 动态导入需要websocket实时推送和回包,分精度计算应该依赖分精度计算程序
     * 静态导入瞬时数据量大, 不适合websocket推送和回包处理,容易导致北向服务处理不过来挂掉,需要走直接插入表的逻辑;
     * 静态导入可以直接生成分精度,同时考虑到分精度服务在同步计算,不修改分精度计算的表号-功能号和项目的的计算进度回拨,只是单纯插入数据
     */
    private Boolean dynamic = false;

    /**
     * 是否读取结束了
     */
    private Boolean readAll = false;
    /**
     * 数据的开始时间
     */
    private String startTime;
    /**
     * 数据的结束时间
     */
    private String endTime;
    /**
     * 最后一次读取的行数
     */
    private Integer lastRowNumber;
    /**
     * 上一次保存数据的时间
     */
    private LocalDateTime latestSaveTime = LocalDateTime.now();
    /**
     * 总处理数据行数
     * （不含表头）
     */
    private AtomicLong totalCount = new AtomicLong();

    private List<String> latestData;

    private Integer sleepMillis = 1000;
    /**
     * 预留，暂时不使用
     * 第几行读取 默认为空，如果又合并的单元格就去获取EasyExcel 内置ReadSheetHolder 然后获取读取的开始行
     * 从没有合并的单元格，则用不上该属性
     */
//    private Integer headRowNumber = null;
    public ExcelListener(Consumer<ExcelData> consumer, Integer batchSize) {
        this.consumer = consumer;
        this.batchSize = batchSize;
    }

    @Override
    public void invoke(T object, AnalysisContext context) {
        lastRowNumber = context.readRowHolder().getRowIndex();
        if (ObjectUtil.isEmpty(buildding)) {
            buildding = context.readSheetHolder().getSheetName();
        }
        if (ObjectUtil.isEmpty(meterList)) {
            meterList = new ArrayList<>(JSON.parseObject(JSON.toJSONString(object, SerializerFeature.WriteMapNullValue), new TypeReference<HashMap<Integer, String>>() {
            }).values());
            return;
        }
        if (ObjectUtil.isEmpty(funcidList)) {
            funcidList = new ArrayList<String>(JSON.parseObject(JSON.toJSONString(object, SerializerFeature.WriteMapNullValue), new TypeReference<HashMap<Integer, String>>() {
            }).values());
            return;
        }
        if (ObjectUtil.isEmpty(descriptionList)) {
            descriptionList = new ArrayList<String>(JSON.parseObject(JSON.toJSONString(object, SerializerFeature.WriteMapNullValue), new TypeReference<HashMap<Integer, String>>() {
            }).values());
            return;
        }
        if (onlyReadHeader) {
            hasNext = false;
            return;
        }
        List<String> data = new ArrayList<String>(JSON.parseObject(JSON.toJSONString(object, SerializerFeature.WriteMapNullValue), new TypeReference<HashMap<Integer, String>>() {
        }).values());

        String dataTime = data.get(0);
        try {
            DateUtil.parse(dataTime);
            if (dynamic && ObjectUtil.isNotEmpty(latestData) && dataTime.compareTo(latestData.get(0)) < 0) {
                log.error("Excel行数:[{}]行数据非单调递增，当前行[{}],上一行[{}]", context.readRowHolder().getRowIndex(), data, latestData);
            }
        } catch (Exception e) {
            log.info("Excel行数:[{}]行数据异常，数据内容:[{}],解析后格式:{}", context.readRowHolder().getRowIndex(), JSON.toJSONString(object), data);
            return;
        }
        latestData = data;
        if (StrUtil.isAllNotEmpty(dataTime, startTime, endTime)) {
            if (startTime.compareTo(dataTime) > 0) {
                //比开始时间还小的数据直接过滤掉
                return;
            }
            if (endTime.compareTo(dataTime) < 0) {
                //比结束时间还大,停止读取
                hasNext = false;
                if (consumer != null) {
                    consumer.accept(ExcelData.builder().dataList(dataList).descriptionList(descriptionList).funcidList(funcidList).meterList(meterList).buildding(buildding).dynamic(dynamic).build());
                    dataList = new ArrayList<>();
                    latestSaveTime = LocalDateTime.now();
                }
                return;
            }
        }
        dataList.add(data);
        if (totalCount.incrementAndGet() % 1000 == 0) {
            log.info("已经处理Excel行数:[{}]行", totalCount.get());
        }
        //批量处理的，数据大于batchSize或者时间超过10秒保存一次
        if (ObjectUtil.isAllNotEmpty(batchSize, consumer) && (dataList.size() >= batchSize || Math.abs(DateUtil.betweenTwoTime(latestSaveTime, LocalDateTime.now(), ChronoUnit.SECONDS)) > 10)) {
            consumer.accept(ExcelData.builder().dataList(dataList).descriptionList(descriptionList).funcidList(funcidList).meterList(meterList).buildding(buildding).dynamic(dynamic).build());
            dataList = new ArrayList<>();
            latestSaveTime = LocalDateTime.now();
            ThreadUtil.safeSleep(sleepMillis);
        }
    }

    @Override
    public void doAfterAllAnalysed(AnalysisContext context) {
        if (consumer != null) {
            consumer.accept(ExcelData.builder().dataList(dataList).descriptionList(descriptionList).funcidList(funcidList).meterList(meterList).buildding(buildding).build());
            dataList = new ArrayList<>();
            latestSaveTime = LocalDateTime.now();
        }
        readAll = true;
        log.warn("所有数据解析完成！");
    }

    /**
     * Verify that there is another piece of data.You can stop the read by returning false
     *
     * @param context
     * @return
     */
    @Override
    public boolean hasNext(AnalysisContext context) {
        return hasNext;
    }

    /**
     * The current method is called when extra information is returned
     *
     * @param extra   extra information
     * @param context analysis context
     */
    @Override
    public void extra(CellExtra extra, AnalysisContext context) {
        super.extra(extra, context);
    }


    // 这个是读取异常是的回调
    @Override
    public void onException(Exception exception, AnalysisContext context) throws Exception {
        // 如果是某一个单元格的转换异常 能获取到具体行号
        if (exception instanceof ExcelDataConvertException) {
            ExcelDataConvertException excelDataConvertException = (ExcelDataConvertException) exception;
            log.error("第{}行，第{}列解析异常", excelDataConvertException.getRowIndex(), excelDataConvertException.getColumnIndex());
        }
        log.error("解析失败，但是继续解析下一行:{} ", exception.getMessage(), exception);
//            super.onException(exception, context);
    }


//    // 这个是读取单元格和并时的信息
//    @Override
//    public void extra(CellExtra extra, AnalysisContext context) {
//        // 解析合并单元格的信息 利用反射给合并的单元格读不到值时，进行赋值
//        if(headRowNumber==null){
//            headRowNumber = context.readSheetHolder().getHeadRowNumber();
//        }
//        // 获取合并后的第一个索引
//        Integer index = extra.getFirstRowIndex() - headRowNumber;
//        // 获取合并后的最后一个索引
//        Integer lastRowIndex = extra.getLastRowIndex() - headRowNumber;
//        // 获取合并后的第一个值
//        T t = excels.get(index);
//        // 利用反射获取所有私有属性
//        Field[] fields = t.getClass().getDeclaredFields();
//        for (Field field:fields) {
//            // 让该属性可操作
//            field.setAccessible(true);
//            // 获取EasyExcel 原有注解
//            ExcelProperty annotation = field.getAnnotation(ExcelProperty.class);
//            // 循环遍历空缺的值
//            for (int i = index+1; i <= lastRowIndex; i++) {
//                // 当读取的列索引等于注解的索引时，表示就是该属性赋值
//                if (extra.getFirstColumnIndex().equals(annotation.index())){
//                    // 利用反射赋值
//                    field.set(excels.get(i),field.get(t));
//                }
//            }
//        }
//
//    }
}
